/*
 *  Copyright (C) 2004 Florian Schirmer <jolt@tuxbox.org>
 *  Copyright (C) 2005 Waldemar Brodkorb <wbx@openwrt.org>
 *  Copyright (C) 2006 Felix Fietkau <nbd@openwrt.org>
 *  Copyright (C) 2006 Michael Buesch <mb@bu3sch.de>
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/types.h>
#include <linux/ssb/ssb.h>
#include <linux/proc_fs.h>
#include <asm/bootinfo.h>
#include <asm/reboot.h>
#include <asm/time.h>
#include <bcm47xx.h>
#include <asm/fw/cfe/cfe_api.h>
#include <linux/time.h>


struct proc_dir_entry *vtk_proc_root;
static struct proc_dir_entry *vtk_proc_env_dir = NULL;

static int vtk_p_proc_env_read(char *page, char **start, off_t pos, int count,
                                     int *eof, void *data);
static int vtk_p_proc_env_write(struct file *file, const char *buffer, 
                                      unsigned long count, void *data);


#define K1SEG	0xa0000000
#define ICS_RESET	(0x18000110|K1SEG)
#define WD_COUNTER	(0x18000080|K1SEG)

struct ssb_bus ssb_bcm47xx;
EXPORT_SYMBOL(ssb_bcm47xx);
EXPORT_SYMBOL(get_uboot_env);
EXPORT_SYMBOL(hal_dev_post_led);

struct timer_list SYSTEM_LED_Timer;
static int hal_system_led = 1;  /* 1:Off */ 

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: hal_system_led_timeout
 * PURPOSE:
 *      To handle POST LED function.
 * PARAMs:
 *		INPUT:  None.
 *		OUTPUT: None.
 * RETURN: None.
 * NOTES:
 *      None.
 *-----------------------------------------------------------------------------
 */
static void hal_system_led_timeout(void)
{
    int                     gpio_val;

    /* Set GPIO #6 from 0 to 1. */
    gpio_val = *(volatile unsigned long *)(0xb8000064);
    if( hal_system_led==1 )
    {
        gpio_val = (gpio_val & ~0x2);   /* Set to 0 to turn On. */
        hal_system_led = !hal_system_led;
    }
    else
    {
        gpio_val = gpio_val | 0x2;     /* Set to 1 to turn Off. */
        hal_system_led = !hal_system_led;
    }
	*(volatile unsigned long *)(0xb8000064) = gpio_val;

    SYSTEM_LED_Timer.expires = jiffies + HZ;
    add_timer(&SYSTEM_LED_Timer);
}

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: hal_system_led_timer_delete
 * PURPOSE:
 *      Delete POST LED timer.
 * PARAMs:
 *		INPUT:  None.
 *		OUTPUT: None.
 * RETURN: None.
 * NOTES:
 *      None.
 *-----------------------------------------------------------------------------
 */
static void hal_system_led_timer_delete( void )
{
    del_timer_sync(&SYSTEM_LED_Timer);
    SYSTEM_LED_Timer.function = NULL;
}

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: hal_system_led_timer_create
 * PURPOSE:
 *      Craete timer for POST LED blinking
 * PARAMs:
 *		INPUT:  None.
 *		OUTPUT: None.
 * RETURN: None.
 * NOTES:
 *      None.
 *-----------------------------------------------------------------------------
 */
static void hal_system_led_timer_create( void )
{
    int                     gpio_val;

    if(SYSTEM_LED_Timer.function == NULL)
    {
        gpio_val = *(volatile unsigned long *)(0xb8000068);

        /* Set GPIO pin 6 to Output mode. */
        gpio_val = gpio_val | 0x2;
        
        *(volatile unsigned long *)(0xb8000068) = gpio_val;
 
        init_timer(&SYSTEM_LED_Timer);
        SYSTEM_LED_Timer.function=hal_system_led_timeout;
        SYSTEM_LED_Timer.expires = jiffies +HZ;
        add_timer(&SYSTEM_LED_Timer);
    }
}

/*-----------------------------------------------------------------------------
 * FUNCTION NAME: hal_dev_post_led
 * PURPOSE:
 *      To set POST LED function
 * PARAMs:
 *		INPUT:  led     -  1:on /0:off /2:blinking
 *		OUTPUT: None.
 * RETURN: None.
 * NOTES:
 *      None.
 *-----------------------------------------------------------------------------
 */
void hal_dev_post_led(unsigned int led)
{
    int                     gpio_val;

    if(led ==0 || led ==1)
    {
        hal_system_led_timer_delete();
        /* Set GPIO #6 from 0 to 1. */
        gpio_val = *(volatile unsigned long *)(0xb8000064);
        if( led==0 )
        {
            gpio_val = (gpio_val & ~0x2);   /* Set to 0 to turn On. */
            hal_system_led = led;
        }
        else
        {
            gpio_val = gpio_val | 0x2;     /* Set to 1 to turn Off. */
            hal_system_led = led;
        }
    	*(volatile unsigned long *)(0xb8000064) = gpio_val;
    }
    else
        hal_system_led_timer_create();

    //printk("%s, %d: Post LED %d!\n",__func__, __LINE__, led); 
}


/*-----------------------------------------------------------------------------
 * FUNCTION NAME: get_uboot_env
 * PURPOSE: Get u-boot environment paramenters.
 *
 * PARAMs:
 *		INPUT:  const char *var,
 *		OUTPUT: char *val
 * RETURN:	None.
 *
 * NOTES: Abble, source from 9528 
 *-----------------------------------------------------------------------------
 */
#define UBOOT_ENV_BASE		0xbc040004
#define __USE_CLIB__		1

void get_uboot_env(const char *var, char *val)
{
    volatile char *env_tmp, *env_ptr;
    char var_str[128];
    int i=1, found=-1;
#ifndef __USE_CLIB__
    int k, j;
#endif
    int var_len, env_sp[100], env_var_n=1;

    env_tmp = (const char *)UBOOT_ENV_BASE;
    env_ptr = env_tmp;
    env_sp[0] = 0;
    /* count number of env_strings */
    while(1) {
        if(*env_ptr == 0) {
            if(*(env_ptr+1) == 0) break;
            else {
                env_sp[env_var_n] = i;
                env_var_n++;
            }
        }
        env_ptr++;
        i++;
    }

#if __USE_CLIB__
    var_len = strlen(var);
#else
    env_ptr = var;
    var_len = 0;
    while(*env_ptr++) {
        var_len++;
    }
#endif
    //var name is too long
    if(var_len > 126) {
        val[0] = 0;
        return;
    }

#if __USE_CLIB__
    memcpy((void *)var_str, (void *)var, var_len);
#else
    env_ptr = var;
    for(j=0; j<var_len; j++) {
        var_str[j] = *env_ptr++;
    }
#endif
    var_str[var_len++] = '=';
    var_str[var_len] = 0;
    for(i=0; i<env_var_n; i++) {
#if __USE_CLIB__
        if(strncmp((char *)(env_tmp +env_sp[i]), var_str, var_len) == 0) {
            found = i;
            break;
        }
#else
        k=0;
        env_ptr = (char *)(env_tmp +env_sp[i]);
        for(j=0; j< var_len; j++) {
           if(*env_ptr++ != var_str[j]) {
               k = -1;
               break;
           }
        }
        if(k == 0) {
            found = i;
            break;
        }
#endif
    }

    if(found == -1) {
        val[0] = 0;
        return;
    }
    else {
        env_ptr = (char *)(env_tmp +env_sp[found] +var_len);
#if __USE_CLIB__
        memcpy((void *)val, (void *)env_ptr, strlen((char *)env_ptr)+1);
#else
        j = 0;
        while(*env_ptr) {
            val[j++] = *env_ptr++;
        }
        val[j] = '\0';
#endif
    }
}



static void bcm47xx_machine_restart(char *command)
{
	printk(KERN_ALERT "Please stand by while rebooting the system...\n");
	local_irq_disable();
	/* Set the watchdog timer to reset immediately */
	ssb_chipco_watchdog_timer_set(&ssb_bcm47xx.chipco, 1);
	while (1)
		cpu_relax();
}

static void bcm47xx_machine_halt(void)
{
	/* Disable interrupts and watchdog and spin forever */
	local_irq_disable();
	ssb_chipco_watchdog_timer_set(&ssb_bcm47xx.chipco, 0);
	while (1)
		cpu_relax();
}

static void str2eaddr(char *str, char *dest)
{
	int i = 0;

	if (str == NULL) {
		memset(dest, 0, 6);
		return;
	}

	for (;;) {
		dest[i++] = (char) simple_strtoul(str, NULL, 16);
		str += 2;
		if (!*str++ || i == 6)
			break;
	}
}

static int bcm47xx_get_invariants(struct ssb_bus *bus,
				   struct ssb_init_invariants *iv)
{
	char buf[100];

	/* Fill boardinfo structure */
	memset(&(iv->boardinfo), 0 , sizeof(struct ssb_boardinfo));

	if (cfe_getenv("boardvendor", buf, sizeof(buf)) >= 0)
		iv->boardinfo.type = (u16)simple_strtoul(buf, NULL, 0);
	if (cfe_getenv("boardtype", buf, sizeof(buf)) >= 0)
		iv->boardinfo.type = (u16)simple_strtoul(buf, NULL, 0);
	if (cfe_getenv("boardrev", buf, sizeof(buf)) >= 0)
		iv->boardinfo.rev = (u16)simple_strtoul(buf, NULL, 0);

	/* Fill sprom structure */
	memset(&(iv->sprom), 0, sizeof(struct ssb_sprom));
	iv->sprom.revision = 3;

	if (cfe_getenv("et0macaddr", buf, sizeof(buf)) >= 0)
		str2eaddr(buf, iv->sprom.et0mac);
	if (cfe_getenv("et1macaddr", buf, sizeof(buf)) >= 0)
		str2eaddr(buf, iv->sprom.et1mac);
	if (cfe_getenv("et0phyaddr", buf, sizeof(buf)) >= 0)
		iv->sprom.et0phyaddr = simple_strtoul(buf, NULL, 10);
	if (cfe_getenv("et1phyaddr", buf, sizeof(buf)) >= 0)
		iv->sprom.et1phyaddr = simple_strtoul(buf, NULL, 10);
	if (cfe_getenv("et0mdcport", buf, sizeof(buf)) >= 0)
		iv->sprom.et0mdcport = simple_strtoul(buf, NULL, 10);
	if (cfe_getenv("et1mdcport", buf, sizeof(buf)) >= 0)
		iv->sprom.et1mdcport = simple_strtoul(buf, NULL, 10);

	return 0;
}

void __init plat_mem_setup(void)
{
	int err;

	err = ssb_bus_ssbbus_register(&ssb_bcm47xx, SSB_ENUM_BASE,
				      bcm47xx_get_invariants);
	if (err)
		panic("Failed to initialize SSB bus (err %d)\n", err);

	_machine_restart = bcm47xx_machine_restart;
	_machine_halt = bcm47xx_machine_halt;
	pm_power_off = bcm47xx_machine_halt;
}


static int vtk_p_proc_env_read(char *page, char **start, off_t pos, int count,
                                     int *eof, void *data)
{
	int len = 0, limit = count-80;
	volatile char *env_tmp, *env_ptr;
	int tmplen;
	int i,j;
	char pval[256],pvar[256];


	env_tmp = (const char *)UBOOT_ENV_BASE;

	/* check region was initialized */
	if( (env_tmp[0] ==0 ) && (env_tmp[1] ==0 ))
		return len;

	if( (env_tmp[0] ==0xff ) && (env_tmp[1] ==0xff ))
		return len;

	i=0;
	
	while(1){

		memset(pvar,0,256);
		memset(pval,0,256);

		/* get line */
		while(env_tmp[i]!='\0'){
			i++;
			/* worst check */
			if(i>256)
				return len;
		}

		//if(env_tmp[i]==0)
		//	break;
		
		/* separate strings */
		j=0;
		while(j<i){
			if(env_tmp[j]=='='){
				strncpy(pvar,env_tmp,j);
				strncpy(pval,(env_tmp+j+1),i-j-1);
				tmplen=strlen(pvar)+strlen(pval)+2;
				break;
			}
			j++;
		}

		if(j>=i)
			break;
		
		if(len + tmplen > limit){
			len = limit;
			break;
		}

		len += sprintf(page + len, "%s\t%s\n", pvar, pval);
		
		env_tmp += i+1;
		i=0;
	}
	
	return len;
}

static int vtk_p_proc_env_write(struct file *file, const char *buffer, 
                                      unsigned long count, void *data)
{
	return count;
}

int vtk_proc_init(void){

	vtk_proc_root = proc_mkdir("vtk",NULL);
  	if(!vtk_proc_root)
     		return -ENOMEM;

   	vtk_proc_env_dir=create_proc_entry("vtk/env", 0644, NULL);
        
	if(vtk_proc_env_dir==NULL )
    {
    	printk(KERN_ERR "Unable to create /proc/%s/%s entry\n", "vtk", "env");
        	return -ENOMEM;
    	}
        
	/* implement read first */
	vtk_proc_env_dir->read_proc =vtk_p_proc_env_read;
    //vtk_proc_env_dir->write_proc=vtk_p_proc_env_write;
    	vtk_proc_env_dir->write_proc=NULL;

	return 0;
}
