/** @file lldp_main.c
 *
 * OpenLLDP Main
 *
 * See LICENSE file for more info.
 * 
 * Authors: Terry Simons (terry.simons@gmail.com)
 *          Jason Peterson (condurre@users.sourceforge.net)
 *
 *******************************************************************/

#include <unistd.h>
#include <strings.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <net/if.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <malloc.h>
#ifdef __FREEBSD__
#include <netinet/in.h>
#include <net/if_dl.h>
#include <net/if_arp.h>
#include <net/if_media.h>
#include <net/ethernet.h>
#include <net/bpf.h>
#endif /* __FREEBSD __ */


#include "lldp_debug.h"
#include "lldp_port.h"
#include "lldp_neighbor.h"
#include "rx_sm.h"
#include "tx_sm.h"
#include "tlv.h"
/*
#include "lldp_linux_framer.h"
#include "lldp_port.h"
#include "lldp_debug.h"
#include "tlv.h"
#include "lldp_neighbor.h"
#include "rx_sm.h"
#include "tx_sm.h"
#include "tlv.h"
#include "tlv_common.h"
#include "framehandler.h"
#include "sasi_cmd.h"
*/


// This is set to argv[0] on startup.
char *program;

struct lci_s lci;


#define IF_NAMESIZE    16
char *iface_list[IF_NAMESIZE];
int iface_filter = 0;   /* boolean */
int process_loopback = 0;

struct lldp_port *lldp_ports = NULL;
int counter=0;

/****************************************
 *
 * Display our usage information.
 *
 ****************************************/
static void usage( void )
{
    debug_printf(DEBUG_EXCESSIVE, "Entering %s():%d\n", __FUNCTION__, __LINE__);
	//debug_printf(DEBUG_NORMAL, "\nlldpd %s\n", VERSION);
	debug_printf(DEBUG_NORMAL, "\nlldpd version\n");
    debug_printf(DEBUG_NORMAL, "(c) Copyright 2002 - 2007 The OpenLLDP Group\n");
    debug_printf(DEBUG_NORMAL, "Licensed under the BSD license.\n\n");
    debug_printf(DEBUG_NORMAL, "This product borrows some code from the Open1X project"
            ". (http://www.open1x.org)\n\n");
    debug_printf(DEBUG_NORMAL, "Usage: %s "
            "[-i <device>] "
            "[-d <debug_level>] "
	        "[-l <med location file>]"
            "[-f] "
            "[-s] "
	        "[-o] "
            "\n", program);

    debug_printf(DEBUG_NORMAL, "\n\n");
    debug_printf(DEBUG_NORMAL, "-i <interface> : Use <interface> for LLDP transactions\n");
    debug_printf(DEBUG_NORMAL, "-d <debug_level/flags> : Set debug verbosity.\n");
    debug_printf(DEBUG_NORMAL, "-l <file> : LLDP-MED Location Identification Configuration File.\n");
    debug_printf(DEBUG_NORMAL, "-f : Run in forground mode.\n");
    debug_printf(DEBUG_NORMAL, "-s : Remove the existing control socket if found.  (Should only be used in system init scripts!)\n");
    debug_printf(DEBUG_NORMAL, "-o : Process LLDP frames on the loopback interface (mainly for testing)\n");
    debug_printf(DEBUG_NORMAL, "\n\n");

    debug_printf(DEBUG_NORMAL, " <debug_level> can be any of : \n");
    debug_printf(DEBUG_NORMAL, "\tA : Enable ALL debug flags.\n");
    debug_printf(DEBUG_NORMAL, "\tc : Enable CONFIG debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\ts : Enable STATE debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\tt : Enable TLV debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\tm : Enable MSAP debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\ti : Enable INT debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\tn : Enable SNMP debug flag.\n");
    debug_printf(DEBUG_NORMAL, "\tx : Enable EXCESSIVE debug flag.\n");
}
/*
static void walk_port_list( void )  // For debug. //
{
    struct lldp_port *lldp_port = lldp_ports;

    while(lldp_port != NULL)
    {
        debug_printf(DEBUG_INT, "Interface structure @ %X\n", lldp_port);
        debug_printf(DEBUG_INT, "\tName: %s\n", lldp_port->if_name);
        debug_printf(DEBUG_INT, "\tIndex: %d\n", lldp_port->if_index);
        debug_printf(DEBUG_INT, "\tMTU: %d\n", lldp_port->mtu);
        debug_printf(DEBUG_INT, "\tMAC: %X:%X:%X:%X:%X:%X\n", lldp_port->source_mac[0]
                                                            , lldp_port->source_mac[1]
                                                            , lldp_port->source_mac[2]
                                                            , lldp_port->source_mac[3]
                                                            , lldp_port->source_mac[4]
                                                            , lldp_port->source_mac[5]); 
        debug_printf(DEBUG_INT, "\tIP: %d.%d.%d.%d\n", lldp_port->source_ipaddr[0]
                                                     , lldp_port->source_ipaddr[1]
                                                     , lldp_port->source_ipaddr[2]
                                                     , lldp_port->source_ipaddr[3]);
        lldp_port = lldp_port->next;
    }
}
*/
/***************************************
 *
 * Trap a segfault, and exit cleanly.
 *
 ***************************************/
static void handle_segfault( void )
{
    fprintf(stderr, "[FATAL] SIGSEGV  (Segmentation Fault)!\n");
    fflush(stderr); fflush(stdout);
    exit(-1);
}

void cleanupLLDP(struct lldp_port *lldp_port)
{
    struct lldp_msap *msap_cache = NULL;

    debug_printf(DEBUG_NORMAL, "[AYBABTU] We Get Signal!\n");

    lldp_port = lldp_ports;
    while(lldp_port != NULL)
    {
        msap_cache = lldp_port->msap_cache;
        while(msap_cache != NULL)
        {
            if(msap_cache->tlv_list != NULL)
	            destroy_tlv_list(&msap_cache->tlv_list);
      
            if(msap_cache->id != NULL) 
	            free(msap_cache->id);
      
            msap_cache = msap_cache->next;
        }

        free(lldp_port->msap_cache);
        lldp_port->msap_cache = NULL;

        if(lldp_port->if_name != NULL)
        {
            debug_printf(DEBUG_NORMAL, "[CLEAN] %s (%d)\n", lldp_port->if_name, lldp_port->if_index);
            tlvCleanupLLDP(lldp_port);
            socketCleanupLLDP(lldp_port);     
        }
        else
            debug_printf(DEBUG_NORMAL, "[ERROR] Interface index %d with name is NULL at: %s():%d\n", lldp_port->if_index, __FUNCTION__, __LINE__);
    
        lldp_port = lldp_port->next;
    
        // Clean the previous node and move up.
        free(lldp_ports);
        lldp_ports = lldp_port;
    }
  
    if(neighbor_local_sd > 0)
        close(neighbor_local_sd);
    unlink(local.sun_path);

    exit(0);
}


static void initializeLLDP( void )
{
    int             if_index=0, ret;
	int             vtk_socket,vtk_mtu;
    sasi_cmd_lldp_t PortStatus;
	unsigned long   linkstlist;
    int             port_adminState;
    struct lldp_port *lldp_port = NULL;
    int changeBefore;

    /* We need to initialize an LLDP port-per interface */
    /* "lldp_port" will be changed to point at the interface currently being serviced */

	/* Get device port link status. */
	ret=lldp_cmd_lldp_port_link_status_get(&linkstlist);
	//debug_printf(DEBUG_NORMAL, "[TINA 000000a] linkstlist=%d, ret=%d\n",linkstlist,ret);

    socketInitializeLLDP_vtk(&vtk_socket,&vtk_mtu);
    for (if_index=BASED_PORT; if_index<=MAXIMUM_PORT; if_index++)//for (if_index=1; if_index<=10 ; if_index++)
    {
		//debug_printf(DEBUG_NORMAL, "========================================================= %d\n",if_index);   

        /* Create our new interface struct */
        lldp_port = malloc(sizeof(struct lldp_port));
	    memset(lldp_port, 0x0, sizeof(struct lldp_port));

        /* Add it to the global list */
        lldp_port->next = lldp_ports;
		lldp_port->socket = vtk_socket;
        lldp_port->mtu = vtk_mtu;
        lldp_port->if_index = if_index;
        lldp_port->if_name = malloc(IF_NAMESIZE);
        if(lldp_port->if_name == NULL)
        {
            free(lldp_port);
            lldp_port = NULL;
            continue;
        }
        snprintf(lldp_port->if_name, IF_NAMESIZE, "Port %d", if_index);
        debug_printf(DEBUG_INT, "%s (index %d) found. Initializing...\n", lldp_port->if_name, lldp_port->if_index);

        // We want the first state to be LLDP_WAIT_PORT_OPERATIONAL, so we'll blank out everything here.
        lldp_port->portEnabled = 1;
	    lldp_port->rx.frame = calloc(1, lldp_port->mtu - 4);
	    lldp_port->tx.frame = calloc(1, (lldp_port->mtu - 2));

        /* Get data from kernel. by Tina */
        lldp_cmd_lldp_local_port_info_get(if_index, &PortStatus);
		memcpy(lldp_port->source_mac, PortStatus.data.tlv_chass_id.mac_addr, 6);
		memcpy(lldp_port->source_ipaddr, PortStatus.data.tlv_management_addr, 4);

        /*debug_printf(DEBUG_NORMAL, "[TINA 222222] MAC=%02x:%02x:%02x:%02x:%02x:%02x, IP=%d.%d.%d.%d.\n",
		lldp_port->source_mac[0],
		lldp_port->source_mac[1],
		lldp_port->source_mac[2],
		lldp_port->source_mac[3],
		lldp_port->source_mac[4],
		lldp_port->source_mac[5],
		lldp_port->source_ipaddr[0],
		lldp_port->source_ipaddr[1],
		lldp_port->source_ipaddr[2],
		lldp_port->source_ipaddr[3]);*/

        debug_printf(DEBUG_EXCESSIVE, "Initializing TX SM for index %d with name %s\n", lldp_port->if_index, lldp_port->if_name);
        lldp_port->tx.state = TX_LLDP_INITIALIZE;
        lldp_port->tx.txTTL = 120;
        txInitializeLLDP(lldp_port);

        debug_printf(DEBUG_EXCESSIVE, "Initializing RX SM for index %d with name %s\n", lldp_port->if_index, lldp_port->if_name);
        lldp_port->rx.state = LLDP_WAIT_PORT_OPERATIONAL;
        rxInitializeLLDP(lldp_port);

        lldp_port->portEnabled  = 0;

        lldp_cmd_lldp_port_adminState_get( if_index, &port_adminState );
        lldp_port->adminState = port_adminState;

        debug_printf(DEBUG_EXCESSIVE, "Initializing TLV subsystem for index %d with name %s\n", lldp_port->if_index, lldp_port->if_name);
        /* Initialize the TLV subsystem for this interface */
        tlvInitializeLLDP(lldp_port);

	    // Send out the first LLDP frame
	    // This allows other devices to see us right when we come up, rather than 
	    // having to wait for a full timer cycle
	    txChangeToState(lldp_port, TX_IDLE);
	    mibConstrInfoLLDPDU(lldp_port);

        if( (0x01<<if_index) & linkstlist )
    		txFrame(lldp_port);

        debug_printf(DEBUG_EXCESSIVE, "Adding index %d with name %s to global port list\n", lldp_port->if_index, lldp_port->if_name);
        /* Reset the global list to point at the top of the list */
        /* We only want to get here if initialization succeeded  */
        lldp_ports = lldp_port;
    }

    /* Don't forget to initialize the TLV validators... */
    initializeTLVFunctionValidators();

    // When we quit, cleanup.
    signal(SIGTERM, cleanupLLDP);
    signal(SIGINT, cleanupLLDP);
    signal(SIGQUIT, cleanupLLDP);
    signal(SIGSEGV, handle_segfault);
}

int main(int argc, char *argv[])
{
    uid_t uid;
    struct timeval timeout;
    struct timeval;
    int fork = 1;
    int op = 0;
    char *theOpts = "i:d:fshl:o";
    int socket_width = 0;
    time_t current_time = 0;
    time_t last_check = 0;
    int result = 0;
    struct lldp_port *lldp_port = NULL;
    sasi_cmd_lldp_t PortStatus;
    int port_adminState, port_linkStatus;
    struct lldp_port lldp_port_common;
    int port_adminState_old;
	int   adminState, val;

    program = argv[0];

    // Process any arguments we were passed in.
    while ((op = getopt(argc, argv, theOpts)) != EOF)
    {
        switch (op)
        {
            case 'd':
    	        // Set the debug level.
    	        if ((atoi(optarg) == 0) && (optarg[0] != '0'))
    	            debug_alpha_set_flags(optarg);
    	        else
    	            debug_set_flags(atoi(optarg));
    	        break;
            case 'i':
    	        iface_filter = 1;
    	        memcpy(iface_list, optarg, strlen(optarg));
    	        iface_list[IF_NAMESIZE - 1] = '\0';
    	        debug_printf(DEBUG_NORMAL, "Using interface %s\n", iface_list);
    	        break;
            case 'l':
    	        debug_printf(DEBUG_NORMAL, "OpenLLDP wasn't compiled with libconfuse support.\n");
    	        exit(1);
    	        break;
            case 'o':
    	        // Turn on the looback interface. :)
    	        process_loopback = 1;
    	        break;
            case 'f':
    	        fork = 0;
    	        break;
            case 's':
    	        unlink(local.sun_path);
    	        break;
            case 'h':
            default:
            	usage();
            	exit(0);
            	break;
        }
    }

    neighbor_local_sd = socket(AF_UNIX, SOCK_STREAM, 0);

    if(neighbor_local_sd < 0)
	    debug_printf(DEBUG_NORMAL, "[Error] Unable to open unix domain socket for client registration!\n");

    local.sun_family = AF_UNIX;

    strcpy(local.sun_path, "/var/run/lldpd.sock");

    //debug_printf(DEBUG_NORMAL, "%s:%d\n", local.sun_path, strlen(local.sun_path));
 
    // Bind to the neighbor list socket.
    result = bind(neighbor_local_sd, (struct sockaddr *)&local, sizeof(local));
    if(result != 0)
        debug_printf(DEBUG_NORMAL, "[Error] Unable to bind to the unix domain socket for client registration!\n");

    result = listen(neighbor_local_sd, 5);
    if(result != 0)
        debug_printf(DEBUG_NORMAL, "[Error] Unable to listen to the unix domain socket for client registration!\n");

    // Set the socket permissions
    if(chmod("/var/run/lldpd.sock", S_IWOTH) != 0)
        debug_printf(DEBUG_NORMAL, "[Error] Unable to set permissions for domain socket!\n");

    /* Needed for select() */
    fd_set readfds;

    // get uid of user executing program. 
    uid = getuid();
    if (uid != 0)
    {
        debug_printf(DEBUG_NORMAL, "You must be running as root to run %s!\n", program);
        exit(0);
    }

	lci.config_file = NULL;

    /* Initialize2 the LLDP subsystem */
    /* This should happen on a per-interface basis */
	initializeLLDP();

    get_sys_desc();
    get_sys_fqdn();

    if (fork)
    {
        if (daemon(0,0) != 0)
            debug_printf(DEBUG_NORMAL, "Unable to daemonize (%m) at: %s():%d\n",__FUNCTION__, __LINE__);
    }

    lldp_port = lldp_ports;
    lldp_port_common.socket=lldp_port->socket;
    lldp_port_common.mtu= lldp_port->mtu;
    lldp_port_common.rx.frame= calloc(1, lldp_port_common.mtu);
    lldp_port_common.rx.recvsize= 0;

    while(1)
    {
		debug_printf(DEBUG_NORMAL, "\n\n\n\n%s(%d):\n" ,__func__,__LINE__);
        
     /* Set up select() */
        FD_ZERO(&readfds);
        FD_SET(lldp_port_common.socket, &readfds);
        socket_width = lldp_port_common.socket;

        // Will be used to tell select how long to wait for...
        timeout.tv_sec = 2;
        timeout.tv_usec = 0;

        // Timeout after 2 second if nothing is ready
        result = select(socket_width+1, &readfds, NULL, NULL, &timeout);
        time(&current_time);
        if(result > 0)
        {
         /* Have received some LLDP packets. */
            debug_printf(DEBUG_INT, "%s,%d: Have received some LLDP packets.\n",__func__,__LINE__);

            if( FD_ISSET(lldp_port_common.socket, &readfds) )
            {
            	
             		/* Get the frame back from the OS-specific frame handler. */				
	                lldp_read( &lldp_port_common );                    
	                if( lldp_port_common.rx.recvsize <= 0 )
	                {
	                    if(errno != EAGAIN && errno != ENETDOWN)
	                        printf("Error:(%d):%s (%s:%d)\n",errno,strerror(errno),__FUNCTION__,__LINE__);
	                }
	                else
	                {
	                    debug_printf(DEBUG_INT, "Got an LLDP frame %d bytes long on %d!\n", lldp_port_common.rx.recvsize, lldp_port_common.if_index);
	                    lldp_port = lldp_ports;
	                    for( ; lldp_port!=NULL ; )
	                    {
	                        if( lldp_port->if_index == lldp_port_common.if_index )
	                        {               
	                        	//lldp_cmd_lldp_state_get(&val);
	                        	lldp_cmd_lldp_port_adminState_get(lldp_port->if_index, &adminState);
								if( (adminState==lldp_enabledRxOnly) || (adminState==lldp_enabledRxTx) )
	                        	{
		                            memcpy( lldp_port->rx.frame, lldp_port_common.rx.frame, lldp_port_common.rx.recvsize);
		                            lldp_port->rx.recvsize=lldp_port_common.rx.recvsize;

		                         /* Mark that we received a frame so the state machine can process it. */
		                            lldp_port->rx.rcvFrame = 1;
		                            rxStatemachineRun(lldp_port);
	                            }
	                            break;                            
	                        }
	                        lldp_port = lldp_port->next;
	                    }
					}
            }
        }
        else
            if( (result == 0) || (current_time > last_check) )  /* Timeout! */
            {
                debug_printf(DEBUG_INT, "%s,%d: Timeout!\n",__func__,__LINE__);

                lldp_port = lldp_ports;
                for( ; lldp_port!=NULL ; )
                {
                    lldp_port->tick = 1;
                    
                    lldp_cmd_lldp_local_port_info_get(lldp_port->if_index, &PortStatus);
      			    memcpy(lldp_port->source_ipaddr, PortStatus.data.tlv_management_addr, 4);

                    //lldp_cmd_port_linkStatus_get( lldp_port->if_index, &port_linkStatus);
                    //lldp_port->portEnabled=port_linkStatus;   /* ????? ݦAF portEnabled Nq. */
    
                    lldp_cmd_lldp_port_adminState_get( lldp_port->if_index, &port_adminState );
                    lldp_port->adminState=port_adminState;

					/*debug_printf(DEBUG_INT, "counter=%d\n",counter);
					if(counter>=21)
                        lldp_port->adminState = 0;// lldp_disabled
                    if (counter>=48)
                        lldp_port->adminState = 3;
                    if (counter>=63)
                        lldp_port->adminState = 2;
                    if (counter>=79)
                        lldp_port->adminState = 1;*/
                        
					if(PortStatus.data.link_state != 0){
                    debug_printf(DEBUG_INT, "0000000000000000000000000000000000000000000000000\n");
                    debug_printf(DEBUG_INT, "%s(%d): portEnabled=%d, adminState=%d\n",__func__,__LINE__,lldp_port->portEnabled,lldp_port->adminState);
					debug_printf(DEBUG_INT, "0000000000000000000000000000000000000000000000000\n");

                    lldp_cmd_lldp_port_adminState_get(lldp_port->if_index, &adminState);
					if( (adminState==lldp_enabledTxOnly) || (adminState==lldp_enabledRxTx) )
                    {
                		txStatemachineRun(lldp_port);
                    }
                    
                	rxStatemachineRun(lldp_port);
                    

                    //counter++;
                    }
                    
                    lldp_port->tick = 0;
                    lldp_port = lldp_port->next;
                }
            }
            else
            {
                if(errno != EINTR)
                    debug_printf(DEBUG_NORMAL, "[ERROR] %s\n", strerror(errno));
            }

        time(&last_check);
    }
    return 0;
}
